"""Utilities.

Miscellaneous utilities.
"""

# Author: Eric Larson
# License: 3-clause BSD

import hashlib
import os
import re
from shutil import move, copyfile
import subprocess

from sphinx.errors import ExtensionError
import sphinx.util

try:
    from sphinx.util.display import status_iterator  # noqa: F401
except Exception:  # Sphinx < 6
    from sphinx.util import status_iterator  # noqa: F401


logger = sphinx.util.logging.getLogger("sphinx-gallery")


def _get_image():
    try:
        from PIL import Image
    except ImportError as exc:  # capture the error for the modern way
        try:
            import Image
        except ImportError:
            raise ExtensionError(
                "Could not import pillow, which is required "
                f"to rescale images (e.g., for thumbnails): {exc}"
            )
    return Image


def scale_image(in_fname, out_fname, max_width, max_height):
    """Scales image centered in image box using `max_width` and `max_height`.

    The same aspect ratio is retained. If `in_fname` == `out_fname` the image can only
    be scaled down.
    """
    # local import to avoid testing dependency on PIL:
    Image = _get_image()
    img = Image.open(in_fname)
    # XXX someday we should just try img.thumbnail((max_width, max_height)) ...
    width_in, height_in = img.size
    scale_w = max_width / float(width_in)
    scale_h = max_height / float(height_in)

    if height_in * scale_w <= max_height:
        scale = scale_w
    else:
        scale = scale_h

    if scale >= 1.0 and in_fname == out_fname:
        return

    width_sc = int(round(scale * width_in))
    height_sc = int(round(scale * height_in))

    # resize the image using resize; if using .thumbnail and the image is
    # already smaller than max_width, max_height, then this won't scale up
    # at all (maybe could be an option someday...)
    try:  # Pillow 9+
        bicubic = Image.Resampling.BICUBIC
    except Exception:
        bicubic = Image.BICUBIC
    img = img.resize((width_sc, height_sc), bicubic)
    # img.thumbnail((width_sc, height_sc), Image.BICUBIC)
    # width_sc, height_sc = img.size  # necessary if using thumbnail

    # insert centered
    thumb = Image.new("RGBA", (max_width, max_height), (255, 255, 255, 0))
    pos_insert = ((max_width - width_sc) // 2, (max_height - height_sc) // 2)
    thumb.paste(img, pos_insert)

    try:
        thumb.save(out_fname)
    except OSError:
        # try again, without the alpha channel (e.g., for JPEG)
        thumb.convert("RGB").save(out_fname)


def optipng(fname, args=()):
    """Optimize a PNG in place.

    Parameters
    ----------
    fname : str
        The filename. If it ends with '.png', ``optipng -o7 fname`` will
        be run. If it fails because the ``optipng`` executable is not found
        or optipng fails, the function returns.
    args : tuple
        Extra command-line arguments, such as ``['-o7']``.
    """
    fname = str(fname)
    if fname.endswith(".png"):
        # -o7 because this is what CPython used
        # https://github.com/python/cpython/pull/8032
        try:
            subprocess.check_call(
                ["optipng"] + list(args) + [fname],
                stdout=subprocess.PIPE,
                stderr=subprocess.PIPE,
            )
        except (subprocess.CalledProcessError, OSError):  # FileNotFoundError
            pass


def _has_optipng():
    try:
        subprocess.check_call(
            ["optipng", "--version"], stdout=subprocess.PIPE, stderr=subprocess.PIPE
        )
    except OSError:  # FileNotFoundError
        return False
    else:
        return True


def get_md5sum(src_file, mode="b"):
    """Returns md5sum of file.

    Parameters
    ----------
    src_file : str
        Filename to get md5sum for.
    mode : 't' or 'b'
        File mode to open file with. When in text mode, universal line endings
        are used to ensure consistency in hashes between platforms.
    """
    if mode == "t":
        kwargs = {"errors": "surrogateescape", "encoding": "utf-8"}
    else:
        kwargs = {}
    with open(src_file, "r" + mode, **kwargs) as src_data:
        src_content = src_data.read()
        if mode == "t":
            src_content = src_content.encode(**kwargs)
        return hashlib.md5(src_content).hexdigest()


def _replace_md5(fname_new, fname_old=None, method="move", mode="b"):
    fname_new = str(fname_new)  # convert possible Path
    assert method in ("move", "copy")
    if fname_old is None:
        assert fname_new.endswith(".new")
        fname_old = os.path.splitext(fname_new)[0]
    replace = True
    if os.path.isfile(fname_old):
        if get_md5sum(fname_old, mode) == get_md5sum(fname_new, mode):
            replace = False
            if method == "move":
                os.remove(fname_new)
        else:
            logger.debug(f"Replacing stale {fname_old} with {fname_new}")
    if replace:
        if method == "move":
            move(fname_new, fname_old)
        else:
            copyfile(fname_new, fname_old)
    assert os.path.isfile(fname_old)


def _has_pypandoc():
    """Check if pypandoc package available."""
    try:
        import pypandoc  # noqa

        # Import error raised only when function called
        version = pypandoc.get_pandoc_version()
    except (ImportError, OSError):
        return None, None
    else:
        return True, version


def _has_graphviz():
    try:
        import graphviz  # noqa F401
    except ImportError as exc:
        logger.info(
            "`graphviz` required for graphical visualization "
            f"but could not be imported, got: {exc}"
        )
        return False
    return True


def _escape_ansi(s):
    """Remove ANSI terminal formatting characters from a string."""
    return re.sub(r"(?:\x1B[@-_]|[\x80-\x9F])[0-?]*[ -/]*[@-~]", "", s)



r"""Sorters for Sphinx-Gallery (sub)sections.

Sorting key functions for gallery subsection folders and section files.
"""

# Created: Sun May 21 20:38:59 2017
# Author: Óscar Nájera
# License: 3-clause BSD

import os
import types

from sphinx.errors import ConfigError

from .gen_rst import extract_intro_and_title
from .py_source_parser import split_code_and_text_blocks


class ExplicitOrder:
    """Sorting key for all gallery subsections.

    All subsections folders must be listed, otherwise an exception is raised.
    However, you can add '*' as a placeholder to the list. All not-listed
    subsection folders will be inserted at the given position and no
    exception is raised.

    Parameters
    ----------
    ordered_list : list, tuple, or :term:`python:generator`
        Hold the paths of each galleries' subsections.

    Raises
    ------
    ValueError
        Wrong input type or Subgallery path missing.
    """

    def __init__(self, ordered_list):
        if not isinstance(ordered_list, (list, tuple, types.GeneratorType)):
            raise ConfigError(
                "ExplicitOrder sorting key takes a list, "
                "tuple or Generator, which hold"
                "the paths of each gallery subfolder"
            )

        self.ordered_list = [
            "*" if path == "*" else os.path.normpath(path) for path in ordered_list
        ]
        try:
            i = ordered_list.index("*")
            self.has_wildcard = True
            self.ordered_list_start = self.ordered_list[:i]
            self.ordered_list_end = self.ordered_list[i + 1 :]
        except ValueError:  # from index("*")
            self.has_wildcard = False
            self.ordered_list_start = []
            self.ordered_list_end = self.ordered_list

    def __call__(self, item):
        """
        Return an integer suited for ordering the items.

        If the ordered_list contains a wildcard "*", items before "*" will return
        negative numbers, items after "*" will have positive numbers, and
        not-listed items will return 0.

        If there is no wildcard, all items with return positive numbers, and
        not-listed items will raise a ConfigError.
        """
        if item in self.ordered_list_start:
            return self.ordered_list_start.index(item) - len(self.ordered_list_start)
        elif item in self.ordered_list_end:
            return self.ordered_list_end.index(item) + 1
        else:
            if self.has_wildcard:
                return 0
            else:
                raise ConfigError(
                    "The subsection folder {!r} was not found in the "
                    "'subsection_order' config. If you use an explicit "
                    "'subsection_order', you must specify all subsection folders "
                    "or add '*' as a wildcard to collect all not-listed subsection "
                    "folders.".format(item)
                )

    def __repr__(self):
        return f"<{self.__class__.__name__} : {self.ordered_list}>"


class _SortKey:
    """Base class for section order key classes."""

    def __init__(self, src_dir):
        self.src_dir = src_dir

    def __repr__(self):
        return f"<{self.__class__.__name__}>"


class NumberOfCodeLinesSortKey(_SortKey):
    """Sort examples in src_dir by the number of code lines.

    Parameters
    ----------
    src_dir : str
        The source directory.
    """

    def __call__(self, filename):
        """Return number of code lines in `filename`."""
        src_file = os.path.normpath(os.path.join(self.src_dir, filename))
        file_conf, script_blocks = split_code_and_text_blocks(src_file)
        amount_of_code = sum(
            [
                len(bcontent)
                for blabel, bcontent, lineno in script_blocks
                if blabel == "code"
            ]
        )
        return amount_of_code


class FileSizeSortKey(_SortKey):
    """Sort examples in src_dir by file size.

    Parameters
    ----------
    src_dir : str
        The source directory.
    """

    def __call__(self, filename):
        """Return file size."""
        src_file = os.path.normpath(os.path.join(self.src_dir, filename))
        return int(os.stat(src_file).st_size)


class FileNameSortKey(_SortKey):
    """Sort examples in src_dir by file name.

    Parameters
    ----------
    src_dir : str
        The source directory.
    """

    def __call__(self, filename):
        """Return `filename`."""
        return filename


class ExampleTitleSortKey(_SortKey):
    """Sort examples in src_dir by example title.

    Parameters
    ----------
    src_dir : str
        The source directory.
    """

    def __call__(self, filename):
        """Return title of example."""
        src_file = os.path.normpath(os.path.join(self.src_dir, filename))
        _, script_blocks = split_code_and_text_blocks(src_file)
        _, title = extract_intro_and_title(src_file, script_blocks[0][1])
        return title


class FunctionSortKey:
    """Sort examples using a function passed through to :py:func:`sorted`.

    Parameters
    ----------
    func : :external+python:term:`callable`
           sorting key function,
           can only take one argument, i.e. lambda func = arg: arg[0] * arg[1]
    r : str, None
        printable representation of object
    """

    def __init__(self, func, r=None):
        self.f = func
        self.r = r

    def __repr__(self):
        return self.r if self.r else "FunctionSortKey"

    def __call__(self, arg):
        """Return func(arg)."""
        return self.f(arg)



"""
======================
Matplotlib alt text
======================

This example tests that the alt text is generated correctly for matplotlib
figures.
"""

import matplotlib.pyplot as plt

fig, axs = plt.subplots(2, 1, constrained_layout=True)
axs[0].plot([1, 2, 3])
axs[0].set_title("subplot 1")
axs[0].set_xlabel("x label")
axs[0].set_ylabel("y lab")
fig.suptitle("This is a\nsup title")

axs[1].plot([2, 3, 4])
axs[1].set_title("subplot 2")
axs[1].set_xlabel("x label")
axs[1].set_ylabel("y label")

plt.show()


# %%
# Several titles.

# sphinx_gallery_thumbnail_number = -1

plt.plot(range(10))

plt.title("Center Title")
plt.title("Left Title", loc="left")
plt.title("Right Title", loc="right")

plt.show()



"""
Matplotlib animation support
============================

Show a Matplotlib animation, which should end up nicely embedded below.

In order to enable support for animations ``'matplotlib_animations'``
must be set to ``True`` in the sphinx gallery
:ref:`configuration <image_scrapers>`.
"""

import numpy as np
import matplotlib.pyplot as plt
import matplotlib.animation as animation

# Adapted from
# https://matplotlib.org/gallery/animation/basic_example.html


def _update_line(num):
    line.set_data(data[..., :num])
    return (line,)


fig, ax = plt.subplots()
data = np.random.RandomState(0).rand(2, 25)
(line,) = ax.plot([], [], "r-")
ax.set(xlim=(0, 1), ylim=(0, 1))
ani = animation.FuncAnimation(fig, _update_line, 25, interval=100, blit=True)



# Author: Chris Holdgraf
# License: 3-clause BSD
"""Binder and Jupyterlite utility functions.

Integration with Binder and Jupyterlite is on an experimental stage. Note that
this API may change in the future.

.. warning::

   Binder is still beta technology, so there may be instability in the
   experience of users who click Binder links.

"""

import os
from pathlib import Path
import shutil
from urllib.parse import quote
import glob
import json

from sphinx.errors import ConfigError
import sphinx.util

from .utils import status_iterator
from . import glr_path_static


logger = sphinx.util.logging.getLogger("sphinx-gallery")


def gen_binder_url(fpath, binder_conf, gallery_conf):
    """Generate a Binder URL according to the configuration in conf.py.

    Parameters
    ----------
    fpath: str
        The path to the `.py` file for which a Binder badge will be generated.
    binder_conf: dict or None
        The Binder configuration dictionary. See `gen_binder_rst` for details.

    Returns
    -------
    binder_url : str
        A URL that can be used to direct the user to the live Binder
        environment.
    """
    # Build the URL
    fpath_prefix = binder_conf["filepath_prefix"]
    link_base = binder_conf["notebooks_dir"]

    # We want to keep the relative path to sub-folders
    relative_link = os.path.relpath(fpath, gallery_conf["src_dir"])
    path_link = os.path.join(link_base, Path(relative_link).with_suffix(".ipynb"))

    # In case our website is hosted in a sub-folder
    if fpath_prefix is not None:
        path_link = "/".join([fpath_prefix.strip("/"), path_link])

    # Make sure we have the right slashes (in case we're on Windows)
    path_link = path_link.replace(os.path.sep, "/")

    # Create the URL
    binder_url = "/".join(
        [
            binder_conf["binderhub_url"],
            "v2",
            "gh",
            binder_conf["org"],
            binder_conf["repo"],
            quote(binder_conf["branch"]),
        ]
    )

    if binder_conf["use_jupyter_lab"] is True:
        binder_url += f"?urlpath=lab/tree/{quote(path_link)}"
    else:
        binder_url += f"?filepath={quote(path_link)}"
    return binder_url


def gen_binder_rst(fpath, binder_conf, gallery_conf):
    """Generate the reST + link for the Binder badge.

    Parameters
    ----------
    fpath: str
        The path to the `.py` file for which a Binder badge will be generated.

    binder_conf: dict or None
        If a dictionary it must have the following keys:

        'binderhub_url'
            The URL of the BinderHub instance that's running a Binder service.
        'org'
            The GitHub organization to which the documentation will be pushed.
        'repo'
            The GitHub repository to which the documentation will be pushed.
        'branch'
            The Git branch on which the documentation exists (e.g., gh-pages).
        'dependencies'
            A list of paths to dependency files that match the Binderspec.

    gallery_conf : dict
        Sphinx-Gallery configuration dictionary.

    Returns
    -------
    rst : str
        The reStructuredText for the Binder badge that links to this file.
    """
    binder_url = gen_binder_url(fpath, binder_conf, gallery_conf)
    # In theory we should be able to use glr_path_static for this, but Sphinx
    # only allows paths to be relative to the build root. On Linux, absolute
    # paths can be used and they work, but this does not seem to be
    # documented behavior:
    #     https://github.com/sphinx-doc/sphinx/issues/7772
    # And in any case, it does not work on Windows, so here we copy the SVG to
    # `images` for each gallery and link to it there. This will make
    # a few copies, and there will be an extra in `_static` at the end of the
    # build, but it at least works...
    physical_path = os.path.join(
        os.path.dirname(fpath), "images", "binder_badge_logo.svg"
    )
    os.makedirs(os.path.dirname(physical_path), exist_ok=True)
    if not os.path.isfile(physical_path):
        shutil.copyfile(
            os.path.join(glr_path_static(), "binder_badge_logo.svg"), physical_path
        )
    rst = (
        "\n"
        "  .. container:: binder-badge\n\n"
        "    .. image:: images/binder_badge_logo.svg\n"
        "      :target: {}\n"
        "      :alt: Launch binder\n"
        "      :width: 150 px\n"
    ).format(binder_url)
    return rst


def copy_binder_files(app, exception):
    """Copy all Binder requirements and notebooks files."""
    if exception is not None:
        return

    if app.builder.name not in ["html", "readthedocs"]:
        return

    gallery_conf = app.config.sphinx_gallery_conf
    binder_conf = gallery_conf["binder"]

    if not len(binder_conf) > 0:
        return

    logger.info("copying binder requirements...", color="white")
    _copy_binder_reqs(app, binder_conf)
    _copy_binder_notebooks(app)


def _copy_binder_reqs(app, binder_conf):
    """Copy Binder requirements files to a "binder" folder in the docs."""
    path_reqs = binder_conf["dependencies"]
    for path in path_reqs:
        if not os.path.exists(os.path.join(app.srcdir, path)):
            raise ConfigError(
                "Couldn't find the Binder requirements file: {}, "
                "did you specify the path correctly?".format(path)
            )

    binder_folder = os.path.join(app.outdir, "binder")
    if not os.path.isdir(binder_folder):
        os.makedirs(binder_folder)

    # Copy over the requirements to the output directory
    for path in path_reqs:
        shutil.copy(os.path.join(app.srcdir, path), binder_folder)


def _remove_ipynb_files(path, contents):
    """Remove files with `.ipynb` and directories named `images` from list of files.

    `contents` should be a list of files, which is returned after file removal.
    Used with the `shutil` "ignore" keyword to filter out non-ipynb files.
    """
    contents_return = []
    for entry in contents:
        if entry.endswith(".ipynb"):
            # Don't include ipynb files
            pass
        elif (entry != "images") and os.path.isdir(os.path.join(path, entry)):
            # Don't include folders not called "images"
            pass
        else:
            # Keep everything else
            contents_return.append(entry)
    return contents_return


def _copy_binder_notebooks(app):
    """Copy Jupyter notebooks to the binder notebooks directory.

    Copy each output gallery directory structure but only including the
    Jupyter notebook files.
    """
    gallery_conf = app.config.sphinx_gallery_conf
    gallery_dirs = gallery_conf["gallery_dirs"]
    binder_conf = gallery_conf["binder"]
    notebooks_dir = os.path.join(app.outdir, binder_conf["notebooks_dir"])
    shutil.rmtree(notebooks_dir, ignore_errors=True)
    os.makedirs(notebooks_dir)

    if not isinstance(gallery_dirs, (list, tuple)):
        gallery_dirs = [gallery_dirs]

    iterator = status_iterator(
        gallery_dirs, "copying binder notebooks...", length=len(gallery_dirs)
    )

    for i_folder in iterator:
        shutil.copytree(
            os.path.join(app.srcdir, i_folder),
            os.path.join(notebooks_dir, i_folder),
            ignore=_remove_ipynb_files,
        )


def check_binder_conf(binder_conf):
    """Check to make sure that the Binder configuration is correct."""
    # Grab the configuration and return None if it's not configured
    binder_conf = {} if binder_conf is None else binder_conf.copy()
    if not isinstance(binder_conf, dict):
        raise ConfigError("`binder_conf` must be a dictionary or None.")
    if len(binder_conf) == 0:
        return binder_conf

    # Ensure all fields are populated
    req_values = ["binderhub_url", "org", "repo", "branch", "dependencies"]
    optional_values = ["filepath_prefix", "notebooks_dir", "use_jupyter_lab"]
    missing_values = []
    for val in req_values:
        if binder_conf.get(val) is None:
            missing_values.append(val)

    if len(missing_values) > 0:
        raise ConfigError(f"binder_conf is missing values for: {missing_values}")

    for key in binder_conf.keys():
        if key not in (req_values + optional_values):
            raise ConfigError(f"Unknown Binder config key: {key}")

    # Ensure we have http in the URL
    if not any(
        binder_conf["binderhub_url"].startswith(ii) for ii in ["http://", "https://"]
    ):
        raise ConfigError(
            "did not supply a valid url, " "gave binderhub_url: {}".format(
                binder_conf["binderhub_url"]
            )
        )

    # Ensure we have at least one dependency file
    # Need at least one of these three files
    required_reqs_files = ["requirements.txt", "environment.yml", "Dockerfile"]
    path_reqs = binder_conf["dependencies"]
    if isinstance(path_reqs, str):
        path_reqs = [path_reqs]
        binder_conf["dependencies"] = path_reqs
    elif not isinstance(path_reqs, (list, tuple)):
        raise ConfigError(
            "`dependencies` value should be a list of strings. " "Got type {}.".format(
                type(path_reqs)
            )
        )

    binder_conf.setdefault("filepath_prefix")
    binder_conf.setdefault("notebooks_dir", "notebooks")
    binder_conf.setdefault("use_jupyter_lab", False)
    path_reqs_filenames = [os.path.basename(ii) for ii in path_reqs]
    if not any(ii in path_reqs_filenames for ii in required_reqs_files):
        raise ConfigError(
            "Did not find one of `requirements.txt` or `environment.yml` "
            'in the "dependencies" section of the binder configuration '
            "for Sphinx-Gallery. A path to at least one of these files "
            "must exist in your Binder dependencies."
        )
    return binder_conf


def pre_configure_jupyterlite_sphinx(app, config):
    """Configure 'jupyterlite_bind_ipynb_suffix' if jupyterlite enabled.

    Connected to "config-inited" event.
    """
    is_jupyterlite_enabled = (
        "jupyterlite_sphinx" in app.extensions
        and config.sphinx_gallery_conf["jupyterlite"] is not None
    )
    if not is_jupyterlite_enabled:
        return

    # Do not use notebooks as sources for the documentation. See
    # https://jupyterlite-sphinx.readthedocs.io/en/latest/configuration.html#disable-the-ipynb-docs-source-binding
    # for more details
    config.jupyterlite_bind_ipynb_suffix = False


def post_configure_jupyterlite_sphinx(app, config):
    """Check SG "jupyterlite" and update Jupyterlite config "jupyterlite_contents".

    Connected to "config-inited" event but lower priority so called after
    `pre_configure_jupyterlite_sphinx`.
    """
    config.sphinx_gallery_conf["jupyterlite"] = check_jupyterlite_conf(
        config.sphinx_gallery_conf["jupyterlite"], app
    )

    if config.sphinx_gallery_conf["jupyterlite"] is None:
        return

    if config.jupyterlite_contents is None:
        config.jupyterlite_contents = []

    # Append to jupyterlite_contents in case they have been set in conf.py
    config.jupyterlite_contents.append(
        config.sphinx_gallery_conf["jupyterlite"]["jupyterlite_contents"]
    )


def create_jupyterlite_contents(app, exception):
    """Create Jupyterlite contents according to "jupyterlite" configuration."""
    from .gen_rst import _get_callables

    if exception is not None:
        return

    if app.builder.name not in ["html", "readthedocs"]:
        return

    gallery_conf = app.config.sphinx_gallery_conf
    is_jupyterlite_enabled = (
        "jupyterlite_sphinx" in app.extensions
        and gallery_conf["jupyterlite"] is not None
    )

    if not is_jupyterlite_enabled:
        return

    logger.info("copying Jupyterlite contents ...", color="white")
    gallery_dirs = gallery_conf["gallery_dirs"]
    contents_dir = gallery_conf["jupyterlite"]["jupyterlite_contents"]

    shutil.rmtree(contents_dir, ignore_errors=True)
    os.makedirs(contents_dir)

    if not isinstance(gallery_dirs, (list, tuple)):
        gallery_dirs = [gallery_dirs]

    iterator = status_iterator(
        gallery_dirs, "Copying Jupyterlite contents ...", length=len(gallery_dirs)
    )

    for i_folder in iterator:
        shutil.copytree(
            os.path.join(app.srcdir, i_folder),
            os.path.join(contents_dir, i_folder),
            ignore=_remove_ipynb_files,
        )

    notebook_modification_function = gallery_conf["jupyterlite"][
        "notebook_modification_function"
    ]

    if notebook_modification_function is None:
        return

    (notebook_modification_function,) = _get_callables(gallery_conf, "jupyterlite")
    notebook_pattern = os.path.join(contents_dir, "**", "*.ipynb")
    notebook_filename_list = sorted(glob.glob(notebook_pattern, recursive=True))

    logger.info("Modifying Jupyterlite notebooks ...", color="white")
    for notebook_filename in notebook_filename_list:
        with open(notebook_filename) as f:
            notebook_content = json.load(f)

        notebook_modification_function(notebook_content, notebook_filename)

        with open(notebook_filename, "w") as f:
            json.dump(notebook_content, f, indent=2)


def gen_jupyterlite_rst(fpath, gallery_conf):
    """Generate the reST + link for the Binder badge.

    Parameters
    ----------
    fpath: str
        The path to the `.py` file for which a JupyterLite badge will be
        generated.

    gallery_conf : dict
        Sphinx-Gallery configuration dictionary.

    Returns
    -------
    rst : str
        The reStructuredText for the JupyterLite badge that links to this file.
    """
    relative_link = os.path.relpath(fpath, gallery_conf["src_dir"])
    notebook_location = relative_link.replace(".py", ".ipynb")
    # Make sure we have the right slashes (in case we're on Windows)
    notebook_location = notebook_location.replace(os.path.sep, "/")

    lite_root_url = os.path.relpath(
        os.path.join(gallery_conf["src_dir"], "lite"), os.path.dirname(fpath)
    )
    # Make sure we have the right slashes (in case we're on Windows)
    lite_root_url = lite_root_url.replace(os.path.sep, "/")

    if gallery_conf["jupyterlite"]["use_jupyter_lab"]:
        lite_root_url += "/lab"
    else:
        lite_root_url += "/retro/notebooks"

    lite_url = f"{lite_root_url}/?path={notebook_location}"

    # Similar work-around for badge file as in
    # gen_binder_rst
    physical_path = os.path.join(
        os.path.dirname(fpath), "images", "jupyterlite_badge_logo.svg"
    )
    os.makedirs(os.path.dirname(physical_path), exist_ok=True)
    if not os.path.isfile(physical_path):
        shutil.copyfile(
            os.path.join(glr_path_static(), "jupyterlite_badge_logo.svg"), physical_path
        )
    rst = (
        "\n"
        "  .. container:: lite-badge\n\n"
        "    .. image:: images/jupyterlite_badge_logo.svg\n"
        "      :target: {}\n"
        "      :alt: Launch JupyterLite\n"
        "      :width: 150 px\n"
    ).format(lite_url)
    return rst


def check_jupyterlite_conf(jupyterlite_conf, app):
    """Return full JupyterLite configuration with defaults."""
    # app=None can happen for testing
    if app is None:
        is_jupyterlite_disabled = True
    else:
        is_jupyterlite_disabled = (
            "jupyterlite_sphinx" not in app.extensions or jupyterlite_conf is None
        )

    if is_jupyterlite_disabled:
        return None

    if not isinstance(jupyterlite_conf, dict):
        raise ConfigError("`jupyterlite_conf` must be a dictionary")

    conf_defaults = {
        "jupyterlite_contents": "jupyterlite_contents",
        "notebook_modification_function": None,
        "use_jupyter_lab": True,
    }

    result = jupyterlite_conf.copy()
    unknown_keys = set(result) - set(conf_defaults)
    if unknown_keys:
        raise ConfigError(
            f"Found some unknown keys in sphinx_gallery_conf['jupyterlite']: "
            f"{sorted(unknown_keys)}. "
            f"Allowed keys are: {list(conf_defaults)}"
        )

    for key, default_value in conf_defaults.items():
        result.setdefault(key, default_value)
    result["jupyterlite_contents"] = os.path.join(
        app.srcdir, result["jupyterlite_contents"]
    )

    return result



# Author: Óscar Nájera
# License: 3-clause BSD
"""Testing the rst files generator."""

import os
import tempfile
import sys

import pytest

from sphinx.errors import ExtensionError
import sphinx_gallery.docs_resolv as sg


def test_embed_code_links_get_data():
    """Test that we can get data for code links."""
    sg._get_data("https://numpy.org/doc/1.18/reference")
    sg._get_data("http://scikit-learn.org/stable/")  # GZip


def test_shelve(tmpdir):
    """Test if shelve can cache and retrieve data after file is deleted."""
    test_string = "test information"
    tmp_cache = str(tmpdir)
    with tempfile.NamedTemporaryFile("w", delete=False) as f:
        f.write(test_string)
    try:
        # recovers data from temporary file and caches it in the shelve
        file_data = sg.get_data(f.name, tmp_cache)
    finally:
        os.remove(f.name)
    # tests recovered data matches
    assert file_data == test_string

    # test if cached data is available after temporary file has vanished
    assert sg.get_data(f.name, tmp_cache) == test_string

    # shelve keys need to be str in python 2, deal with unicode input
    if sys.version_info[0] == 2:
        assert sg.get_data(f.name, tmp_cache) == test_string


def test_parse_sphinx_docopts():
    data = """
    <script type="text/javascript">
      var DOCUMENTATION_OPTIONS = {
        URL_ROOT:    './',
        VERSION:     '2.0.2',
        COLLAPSE_INDEX: false,
        FILE_SUFFIX: '.html',
        HAS_SOURCE:  true,
        SOURCELINK_SUFFIX: '.txt'
      };
    </script>
    """
    assert sg.parse_sphinx_docopts(data) == {
        "URL_ROOT": "./",
        "VERSION": "2.0.2",
        "COLLAPSE_INDEX": False,
        "FILE_SUFFIX": ".html",
        "HAS_SOURCE": True,
        "SOURCELINK_SUFFIX": ".txt",
    }

    data_sphinx_175 = """
    <script type="text/javascript">
      var DOCUMENTATION_OPTIONS = {
        URL_ROOT: document.getElementById("documentation_options")\
                  .getAttribute('data-url_root'),
        VERSION:     '2.0.2',
        COLLAPSE_INDEX: false,
        FILE_SUFFIX: '.html',
        HAS_SOURCE:  true,
        SOURCELINK_SUFFIX: '.txt'
      };
    </script>
    """
    assert sg.parse_sphinx_docopts(data_sphinx_175) == {
        "VERSION": "2.0.2",
        "COLLAPSE_INDEX": False,
        "FILE_SUFFIX": ".html",
        "HAS_SOURCE": True,
        "SOURCELINK_SUFFIX": ".txt",
    }

    with pytest.raises(ExtensionError):
        sg.parse_sphinx_docopts("empty input")

    with pytest.raises(ExtensionError):
        sg.parse_sphinx_docopts("DOCUMENTATION_OPTIONS = ")

    with pytest.raises(ExtensionError):
        sg.parse_sphinx_docopts("DOCUMENTATION_OPTIONS = {")



r"""Parser for python source files."""

# Created Sun Nov 27 14:03:07 2016
# Author: Óscar Nájera

import codecs
import ast
from io import BytesIO
import re
import tokenize
from textwrap import dedent

from sphinx.errors import ExtensionError
from sphinx.util.logging import getLogger

logger = getLogger("sphinx-gallery")

SYNTAX_ERROR_DOCSTRING = """
SyntaxError
===========

Example script with invalid Python syntax
"""

# The pattern for in-file config comments is designed to not greedily match
# newlines at the start and end, except for one newline at the end. This
# ensures that the matched pattern can be removed from the code without
# changing the block structure; i.e. empty newlines are preserved, e.g. in
#
#     a = 1
#
#     # sphinx_gallery_thumbnail_number = 2
#
#     b = 2
FLAG_START = r"^[\ \t]*#\s*"
FLAG_BODY = r"sphinx_gallery_([A-Za-z0-9_]+)(\s*=\s*(.+))?[\ \t]*\n?"
INFILE_CONFIG_PATTERN = re.compile(FLAG_START + FLAG_BODY, re.MULTILINE)

START_IGNORE_FLAG = FLAG_START + "sphinx_gallery_start_ignore"
END_IGNORE_FLAG = FLAG_START + "sphinx_gallery_end_ignore"
IGNORE_BLOCK_PATTERN = re.compile(
    rf"{START_IGNORE_FLAG}(?:[\s\S]*?){END_IGNORE_FLAG}\n?", re.MULTILINE
)


def parse_source_file(filename):
    """Parse source file into AST node.

    Parameters
    ----------
    filename : str
        File path

    Returns
    -------
    node : AST node
    content : utf-8 encoded string
    """
    with codecs.open(filename, "r", "utf-8") as fid:
        content = fid.read()
    # change from Windows format to UNIX for uniformity
    content = content.replace("\r\n", "\n")

    try:
        node = ast.parse(content)
        return node, content
    except SyntaxError:
        return None, content


def _get_docstring_and_rest(filename):
    """Separate ``filename`` content between docstring and the rest.

    Strongly inspired from ast.get_docstring.

    Returns
    -------
    docstring : str
        docstring of ``filename``
    rest : str
        ``filename`` content without the docstring
    lineno : int
        The line number.
    node : ast.Module
        The ast node. When `filename` parsed with `mode='exec'` node should be
        of type `ast.Module`.
    """
    node, content = parse_source_file(filename)

    if node is None:
        return SYNTAX_ERROR_DOCSTRING, content, 1, node

    if not isinstance(node, ast.Module):
        raise ExtensionError(
            "This function only supports modules. " "You provided {}".format(
                node.__class__.__name__
            )
        )
    if not (
        node.body
        and isinstance(node.body[0], ast.Expr)
        and isinstance(node.body[0].value, ast.Constant)
    ):
        raise ExtensionError(
            'Could not find docstring in file "{}". '
            "A docstring is required by sphinx-gallery "
            'unless the file is ignored by "ignore_pattern"'.format(filename)
        )

    # Python 3.7+ way
    docstring = ast.get_docstring(node)
    assert docstring is not None  # should be guaranteed above
    # This is just for backward compat
    if node.body[0].value.value[:1] == "\n":
        # just for strict backward compat here
        docstring = "\n" + docstring
    ts = tokenize.tokenize(BytesIO(content.encode()).readline)
    # find the first string according to the tokenizer and get its end row
    for tk in ts:
        if tk.exact_type == 3:
            lineno, _ = tk.end
            break
    else:
        lineno = 0

    # This get the content of the file after the docstring last line
    # Note: 'maxsplit' argument is not a keyword argument in python2
    rest = "\n".join(content.split("\n")[lineno:])
    lineno += 1
    return docstring, rest, lineno, node


def extract_file_config(content):
    """Pull out the file-specific config specified in the docstring."""
    file_conf = {}
    for match in re.finditer(INFILE_CONFIG_PATTERN, content):
        name = match.group(1)
        value = match.group(3)
        if value is None:  # a flag rather than a config setting
            continue
        try:
            value = ast.literal_eval(value)
        except (SyntaxError, ValueError):
            logger.warning(
                "Sphinx-gallery option %s was passed invalid value %s", name, value
            )
        else:
            file_conf[name] = value
    return file_conf


def split_code_and_text_blocks(source_file, return_node=False):
    """Return list with source file separated into code and text blocks.

    Parameters
    ----------
    source_file : str
        Path to the source file.
    return_node : bool
        If True, return the ast node.

    Returns
    -------
    file_conf : dict
        File-specific settings given in source file comments as:
        ``# sphinx_gallery_<name> = <value>``
    blocks : list
        (label, content, line_number)
        List where each element is a tuple with the label ('text' or 'code'),
        the corresponding content string of block and the leading line number.
    node : ast.Module
        The parsed ast node.
    """
    docstring, rest_of_content, lineno, node = _get_docstring_and_rest(source_file)
    blocks = [("text", docstring, 1)]

    file_conf = extract_file_config(rest_of_content)

    pattern = re.compile(
        r"(?P<header_line>^#{20,}.*|^# ?%%.*)\s(?P<text_content>(?:^#.*\s?)*)",
        flags=re.M,
    )
    sub_pat = re.compile("^#", flags=re.M)

    pos_so_far = 0
    for match in re.finditer(pattern, rest_of_content):
        code_block_content = rest_of_content[pos_so_far : match.start()]
        if code_block_content.strip():
            blocks.append(("code", code_block_content, lineno))
        lineno += code_block_content.count("\n")

        lineno += 1  # Ignored header line of hashes.
        text_content = match.group("text_content")
        text_block_content = dedent(re.sub(sub_pat, "", text_content)).lstrip()
        if text_block_content.strip():
            blocks.append(("text", text_block_content, lineno))
        lineno += text_content.count("\n")

        pos_so_far = match.end()

    remaining_content = rest_of_content[pos_so_far:]
    if remaining_content.strip():
        blocks.append(("code", remaining_content, lineno))

    out = (file_conf, blocks)
    if return_node:
        out += (node,)
    return out


def remove_ignore_blocks(code_block):
    """
    Return the content of *code_block* with ignored areas removed.

    An ignore block starts with # sphinx_gallery_start_ignore, and ends with
    # sphinx_gallery_end_ignore. These lines and anything in between them will
    be removed, but surrounding empty lines are preserved.

    Parameters
    ----------
    code_block : str
        A code segment.
    """
    num_start_flags = len(re.findall(START_IGNORE_FLAG, code_block, re.MULTILINE))
    num_end_flags = len(re.findall(END_IGNORE_FLAG, code_block, re.MULTILINE))

    if num_start_flags != num_end_flags:
        raise ExtensionError(
            'All "sphinx_gallery_start_ignore" flags must have a matching '
            '"sphinx_gallery_end_ignore" flag!'
        )
    return re.subn(IGNORE_BLOCK_PATTERN, "", code_block)[0]


def remove_config_comments(code_block):
    """
    Return the content of *code_block* with in-file config comments removed.

    Comment lines of the pattern '# sphinx_gallery_[option] = [val]' are
    removed, but surrounding empty lines are preserved.

    Parameters
    ----------
    code_block : str
        A code segment.
    """
    parsed_code, _ = re.subn(INFILE_CONFIG_PATTERN, "", code_block)
    return parsed_code



"""
======================
Link to other packages
======================

Use :mod:`sphinx_gallery` to link to other packages, like
:mod:`numpy`, :mod:`matplotlib.colors`, and :mod:`matplotlib.pyplot`.

FYI this gallery uses :obj:`sphinx_gallery.sorting.FileNameSortKey`.
"""

from warnings import warn

import numpy as np
from matplotlib.colors import is_color_like
from matplotlib.figure import Figure
from itertools import compress  # noqa
import matplotlib
import matplotlib.pyplot as plt
import sphinx_gallery.backreferences

from local_module import N  # N = 1000

t = np.arange(N) / float(N)
win = np.hanning(N)
print(is_color_like("r"))
fig, ax = plt.subplots()
ax.plot(t, win, color="r")
ax.text(0, 1, "png", size=40, va="top")
fig.tight_layout()
orig_dpi = 80.0 if matplotlib.__version__[0] < "2" else 100.0
assert plt.rcParams["figure.dpi"] == orig_dpi
plt.rcParams["figure.dpi"] = 70.0
assert plt.rcParams["figure.dpi"] == 70.0
listy = [0, 1]
compress("abc", [0, 0, 1])
warn("This warning should show up in the output", RuntimeWarning)
x = Figure()  # plt.Figure should be decorated (class), x shouldn't (inst)
# nested resolution resolves to numpy.random.mtrand.RandomState:
rng = np.random.RandomState(0)
# test Issue 583
sphinx_gallery.backreferences.identify_names(
    [("text", "Text block", 1)],
    sphinx_gallery.backreferences._make_ref_regex(),
)
# 583: methods don't link properly
dc = sphinx_gallery.backreferences.DummyClass()
dc.run()
print(dc.prop)



"""Testing sphinx_gallery.load_style extension."""

import os
import pytest


@pytest.mark.conf_file(extensions=["sphinx_gallery.load_style"])
def test_load_style(sphinx_app_wrapper):
    """Testing that style loads properly."""
    sphinx_app = sphinx_app_wrapper.build_sphinx_app()
    cfg = sphinx_app.config
    assert cfg.project == "Sphinx-Gallery <Tests>"
    build_warn = sphinx_app._warning.getvalue()
    assert build_warn == ""
    index_html = os.path.join(sphinx_app_wrapper.outdir, "index.html")
    assert os.path.isfile(index_html)
    with open(index_html) as fid:
        content = fid.read()
    assert (
        'link rel="stylesheet" type="text/css" href="_static/sg_gallery.css' in content
    )



"""
Test without __future__ imports
-------------------------------

Test that __future__ imports inside sphinx_gallery modules does not affect the
parsing of this script.
"""

import sys

PY3_OR_LATER = sys.version_info[0] >= 3

# SyntaxError on Python 2
print(3 / 2, end="")

# Need to make this example fail on Python 3 as well (currently no way to say
# that an example is expected to fail only on Python 2)
if PY3_OR_LATER:
    raise RuntimeError("Forcing this example to fail on Python 3")



# Author: Arturo Amor
# License: 3-clause BSD
"""Test the example recommender plugin."""

import codecs
import os
import re

import pytest

import sphinx_gallery.gen_rst as sg
from sphinx_gallery.recommender import ExampleRecommender, _write_recommendations


def test_example_recommender_methods():
    """Test dict_vectorizer and compute_tf_idf methods."""
    np = pytest.importorskip("numpy")
    recommender = ExampleRecommender()
    D = [{"foo": 1, "bar": 2}, {"foo": 3, "baz": 1}]

    X = recommender.dict_vectorizer(D)
    expected_X = np.array([[2.0, 0.0, 1.0], [0.0, 1.0, 3.0]])
    np.testing.assert_array_equal(X, expected_X)

    X_tfidf = recommender.compute_tf_idf(X)
    expected_X_tfidf = np.array(
        [[0.94215562, 0.0, 0.33517574], [0.0, 0.42423963, 0.90554997]]
    )
    np.testing.assert_array_almost_equal(X_tfidf, expected_X_tfidf)


def test_recommendation_files(gallery_conf):
    """Test generated files and that recommendations are relevant."""
    pytest.importorskip("numpy")
    gallery_conf["recommender"].update(
        [("enable", True), ("rubric_header", "Custom header")]
    )
    file_dict = {
        "fox_jumps_dog.py": "The quick brown fox jumped over the lazy dog",
        "dog_sleeps.py": "The lazy dog slept all day",
        "fox_eats_dog_food.py": "The quick brown fox ate the lazy dog's food",
        "dog_jumps_fox.py": "The quick dog jumped over the lazy fox",
    }

    for file_name, content in file_dict.items():
        file_path = os.path.join(gallery_conf["gallery_dir"], file_name)
        with open(file_path, "w") as f:
            f.write(content)
        sg.save_rst_example("example_rst", file_path, 0, 0, gallery_conf)

        test_file = re.sub(r"\.py$", ".rst", file_path)
        recommendation_file = re.sub(r"\.py$", ".recommendations", file_name)
        with codecs.open(test_file) as f:
            rst = f.read()

        assert recommendation_file in rst

    py_files = [
        fname
        for fname in os.listdir(gallery_conf["gallery_dir"])
        if os.path.splitext(fname)[1] == ".py"
    ]
    gallery_py_files = [
        os.path.join(gallery_conf["gallery_dir"], fname) for fname in py_files
    ]
    recommender = ExampleRecommender(n_examples=1, min_df=1)
    recommender.fit(gallery_py_files)
    recommended_example = recommender.predict(file_path)  # dog_jumps_fox.py

    assert os.path.basename(recommended_example[0]) == "fox_jumps_dog.py"

    # _write_recommendations needs a thumbnail, for writing the
    # `_thumbnail_div` we then create a blank png
    thumb_path = os.path.join(gallery_conf["gallery_dir"], "images/thumb")
    os.makedirs(thumb_path, exist_ok=True)
    png_file = "sphx_glr_fox_jumps_dog_thumb.png"
    png_file_path = os.path.join(thumb_path, png_file)
    with open(png_file_path, "wb") as f:
        b"\x89PNG\r\n\x1a\n"  # generic png file signature

    recommendation_file = re.sub(r"\.py$", ".recommendations", file_path)
    _write_recommendations(recommender, file_path, gallery_conf)
    with codecs.open(recommendation_file) as f:
        rst = f.read()
    assert ".. rubric:: Custom header" in rst



"""
Example with SyntaxError
========================

Sphinx-Gallery uses Python's AST parser, thus you need to have written
valid python code for Sphinx-Gallery to parse it. If your script has a
SyntaxError you'll be presented the traceback and the original code.
"""
# Code source: Óscar Nájera
# License: BSD 3 clause

Invalid Python code
